@@ -1,5 +1,5 @@
 /* 
-Copyright (C) 1993 Free Software Foundation
+Copyright (C) 1993, 1995 Free Software Foundation
 
 This file is part of the GNU IO Library.  This library is free
 software; you can redistribute it and/or modify it under the
@@ -13,8 +13,8 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
-along with GNU CC; see the file COPYING.  If not, write to
-the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+along with this library; see the file COPYING.  If not, write to the Free
+Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 As a special exception, if you link this library with files
 compiled with a GNU compiler to produce an executable, this does not cause
@@ -25,11 +25,62 @@ the executable file might be covered by the GNU General Public License. */
 /* Generic or default I/O operations. */
 
 #include "libioP.h"
+#ifdef __STDC__
 #include <stdlib.h>
+#endif
 #include <string.h>
 
-void _IO_un_link(fp)
-     _IO_FILE *fp;
+void _IO_un_link( _IO_FILE *);
+void _IO_link_in( _IO_FILE *);
+_IO_size_t _IO_least_marker( register _IO_FILE * );
+void _IO_switch_to_main_get_area( _IO_FILE * );
+void _IO_switch_to_backup_area( register _IO_FILE * );
+void _IO_free_backup_area( register _IO_FILE * );
+int __overflow( _IO_FILE *, int );
+static int save_for_backup( _IO_FILE * );
+int __underflow( _IO_FILE * );
+int __uflow( _IO_FILE * );
+void _IO_setb( _IO_FILE *, char *, char *, int );
+void _IO_doallocbuf( register _IO_FILE * );
+int _IO_default_underflow( _IO_FILE * );
+int _IO_default_uflow( _IO_FILE * );
+_IO_size_t _IO_default_xsputn( register _IO_FILE *, const void *, _IO_size_t );
+_IO_size_t _IO_sgetn( _IO_FILE *, void *, _IO_size_t );
+_IO_size_t _IO_default_xsgetn( _IO_FILE *, void *, _IO_size_t );
+int _IO_sync( register _IO_FILE * );
+_IO_FILE* _IO_default_setbuf( register _IO_FILE *, char *, _IO_ssize_t );
+_IO_pos_t _IO_default_seekpos( _IO_FILE *, _IO_pos_t, int );
+int _IO_default_doallocate( _IO_FILE * );
+void _IO_init( register _IO_FILE *, int );
+int _IO_default_sync( _IO_FILE * );
+void _IO_default_finish( _IO_FILE * , int);
+_IO_pos_t _IO_default_seekoff( register _IO_FILE *, _IO_off_t, int, int );
+int _IO_sputbackc( register _IO_FILE *, int );
+int _IO_sungetc( register _IO_FILE * );
+int _IO_set_column( register _IO_FILE *, int );
+unsigned _IO_adjust_column( unsigned, const char *, int );
+int _IO_get_column( register _IO_FILE * );
+int _IO_flush_all( void );
+void _IO_flush_all_linebuffered( void );
+void _IO_unbuffer_all( void );
+void _IO_cleanup( void );
+void _IO_init_marker( struct _IO_marker *, _IO_FILE * );
+void _IO_remove_marker( register struct _IO_marker * );
+int _IO_marker_difference( struct _IO_marker *, struct _IO_marker * );
+int _IO_marker_delta( struct _IO_marker * );
+int _IO_seekmark( _IO_FILE *, struct _IO_marker *, int );
+void _IO_unsave_markers( register _IO_FILE * );
+int _IO_nobackup_pbackfail( register _IO_FILE *, int );
+int _IO_default_pbackfail( register _IO_FILE *, int );
+_IO_pos_t _IO_default_seek( _IO_FILE *, _IO_off_t, int );
+int _IO_default_stat( _IO_FILE *, void * );
+_IO_ssize_t _IO_default_read( register _IO_FILE *, void *, _IO_ssize_t );
+_IO_ssize_t _IO_default_write( register _IO_FILE *, const void *,
+										_IO_ssize_t );
+
+void
+DEFUN(_IO_un_link, (fp),
+      _IO_FILE *fp)
 {
   if (fp->_flags & _IO_LINKED) {
     _IO_FILE **f;
@@ -43,8 +94,9 @@ void _IO_un_link(fp)
   }
 }
 
-void _IO_link_in(fp)
-     _IO_FILE *fp;
+void
+DEFUN(_IO_link_in, (fp),
+      _IO_FILE *fp)
 {
     if ((fp->_flags & _IO_LINKED) == 0) {
 	fp->_flags |= _IO_LINKED;
@@ -57,8 +109,8 @@ void _IO_link_in(fp)
    Assumes the current get area is the main get area. */
 
 _IO_size_t
-_IO_least_marker(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_least_marker, (fp),
+      register _IO_FILE *fp)
 {
   _IO_ssize_t least_so_far = fp->_IO_read_end - fp->_IO_read_base;
   register struct _IO_marker *mark;
@@ -71,8 +123,8 @@ _IO_least_marker(fp)
 /* Switch current get area from backup buffer to (start of) main get area. */
 
 void
-_IO_switch_to_main_get_area(fp)
-     _IO_FILE *fp;
+DEFUN(_IO_switch_to_main_get_area, (fp),
+      _IO_FILE *fp)
 {
   char *tmp;
   fp->_flags &= ~_IO_IN_BACKUP;
@@ -86,8 +138,8 @@ _IO_switch_to_main_get_area(fp)
 /* Switch current get area from main get area to (end of) backup area. */
 
 void
-_IO_switch_to_backup_area(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_switch_to_backup_area, (fp),
+     register _IO_FILE *fp)
 {
   char *tmp;
   fp->_flags |= _IO_IN_BACKUP;
@@ -99,14 +151,14 @@ _IO_switch_to_backup_area(fp)
 }
 
 int
-_IO_switch_to_get_mode(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_switch_to_get_mode, (fp),
+     register _IO_FILE *fp)
 {
   if (fp->_IO_write_ptr > fp->_IO_write_base)
-    if (fp->_jumps->__overflow(fp, EOF) == EOF)
+    if (_IO_OVERFLOW (fp, EOF) == EOF)
       return EOF;
   if (_IO_in_backup(fp))
-    fp->_IO_read_base = fp->_aux_limit;
+    fp->_IO_read_base = fp->_IO_backup_base;
   else
     {
       fp->_IO_read_base = fp->_IO_buf_base;
@@ -122,21 +174,21 @@ _IO_switch_to_get_mode(fp)
 }
 
 void
-_IO_free_backup_area(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_free_backup_area, (fp),
+     register _IO_FILE *fp)
 {
   if (_IO_in_backup (fp))
     _IO_switch_to_main_get_area(fp);  /* Just in case. */
   free (fp->_IO_save_base);
   fp->_IO_save_base = NULL;
   fp->_IO_save_end = NULL;
-  fp->_aux_limit = NULL;
+  fp->_IO_backup_base = NULL;
 }
 
 #if 0
 int
-_IO_switch_to_put_mode(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_switch_to_put_mode, (fp),
+      register _IO_FILE *fp)
 {
   fp->_IO_write_base = fp->_IO_read_ptr;
   fp->_IO_write_ptr = fp->_IO_read_ptr;
@@ -151,16 +203,79 @@ _IO_switch_to_put_mode(fp)
 }
 #endif
 
-int  __overflow(f, ch)
-     _IO_FILE *f;
-     int ch;
+int
+DEFUN(__overflow, (f, ch),
+      _IO_FILE *f AND int ch)
 {
-  return f->_jumps->__overflow(f, ch);
+  return _IO_OVERFLOW (f, ch);
+}
+
+static int
+DEFUN(save_for_backup, (fp),
+      _IO_FILE *fp)
+{
+  /* Append [_IO_read_base.._IO_read_end] to backup area. */
+  int least_mark = _IO_least_marker(fp);
+  /* needed_size is how much space we need in the backup area. */
+  int needed_size = (fp->_IO_read_end - fp->_IO_read_base) - least_mark;
+  int current_Bsize = fp->_IO_save_end - fp->_IO_save_base;
+  int avail; /* Extra space available for future expansion. */
+  int delta;
+  struct _IO_marker *mark;
+  if (needed_size > current_Bsize)
+    {
+      char *new_buffer;
+      avail = 100;
+      new_buffer = (char*)malloc(avail+needed_size);
+      if (new_buffer == NULL)
+	return EOF;		/* FIXME */
+      if (least_mark < 0)
+	{
+	  memcpy(new_buffer + avail,
+		 fp->_IO_save_end + least_mark,
+		 -least_mark);
+	  memcpy(new_buffer +avail - least_mark,
+		 fp->_IO_read_base,
+		 fp->_IO_read_end - fp->_IO_read_base);
+	}
+      else
+	memcpy(new_buffer + avail,
+	       fp->_IO_read_base + least_mark,
+	       needed_size);
+      if (fp->_IO_save_base)
+	free (fp->_IO_save_base);
+      fp->_IO_save_base = new_buffer;
+      fp->_IO_save_end = new_buffer + avail + needed_size;
+    }
+  else
+    {
+      avail = current_Bsize - needed_size;
+      if (least_mark < 0)
+	{
+	  memmove(fp->_IO_save_base + avail,
+		  fp->_IO_save_end + least_mark,
+		  -least_mark);
+	  memcpy(fp->_IO_save_base + avail - least_mark,
+		 fp->_IO_read_base,
+		 fp->_IO_read_end - fp->_IO_read_base);
+	}
+      else if (needed_size > 0)
+	memcpy(fp->_IO_save_base + avail,
+	       fp->_IO_read_base + least_mark,
+	       needed_size);
+    }
+  /* FIXME: Dubious arithmetic if pointers are NULL */
+  fp->_IO_backup_base = fp->_IO_save_base + avail;
+  /* Adjust all the streammarkers. */
+  delta = fp->_IO_read_end - fp->_IO_read_base;
+  for (mark = fp->_markers; mark != NULL; mark = mark->_next)
+    mark->_pos -= delta;
+  return 0;
 }
 
 int
-__underflow(fp)
-     _IO_FILE *fp;
+DEFUN(__underflow, (fp),
+      _IO_FILE *fp)
 {
   if (_IO_in_put_mode(fp))
     if (_IO_switch_to_get_mode(fp) == EOF) return EOF;
@@ -174,76 +289,44 @@ __underflow(fp)
     }
   if (_IO_have_markers(fp))
     {
-      /* Append [_gbase.._IO_read_end] to backup area. */
-      int least_mark = _IO_least_marker(fp);
-      /* needed_size is how much space we need in the backup area. */
-      int needed_size = (fp->_IO_read_end - fp->_IO_read_base) - least_mark;
-      int current_Bsize = fp->_IO_save_end - fp->_IO_save_base;
-      int avail; /* Extra space available for future expansion. */
-      int delta;
-      struct _IO_marker *mark;
-      if (needed_size > current_Bsize)
-	{
-	  char *new_buffer;
-	  avail = 100;
-	  new_buffer = (char*)malloc(avail+needed_size);
-	  if (new_buffer == NULL)
-	    return EOF; /* FIXME */
-	  if (least_mark < 0)
-	    {
-	      memcpy(new_buffer + avail,
-		     fp->_IO_save_end + least_mark,
-		     -least_mark);
-	      memcpy(new_buffer +avail - least_mark,
-		     fp->_IO_read_base,
-		     fp->_IO_read_end - fp->_IO_read_base);
-	    }
-	  else
-	    memcpy(new_buffer + avail,
-		   fp->_IO_read_base + least_mark,
-		   needed_size);
-	  free (fp->_IO_save_base);
-	  fp->_IO_save_base = new_buffer;
-	  fp->_IO_save_end = new_buffer + avail + needed_size;
-	}
-      else
-	{
-	  avail = current_Bsize - needed_size;
-	  if (least_mark < 0)
-	    {
-	      memmove(fp->_IO_save_base + avail,
-		      fp->_IO_save_end + least_mark,
-		      -least_mark);
-	      memcpy(fp->_IO_save_base + avail - least_mark,
-		     fp->_IO_read_base,
-		     fp->_IO_read_end - fp->_IO_read_base);
-	    }
-	  else if (needed_size > 0)
-	    memcpy(fp->_IO_save_base + avail,
-		   fp->_IO_read_base + least_mark,
-		   needed_size);
-	}
-      /* FIXME: Dubious arithmetic if pointers are NULL */
-      fp->_aux_limit = fp->_IO_save_base + avail;
-      /* Adjust all the streammarkers. */
-      delta = fp->_IO_read_end - fp->_IO_read_base;
-      for (mark = fp->_markers; mark != NULL; mark = mark->_next)
-	mark->_pos -= delta;
+      if (save_for_backup (fp))
+	return EOF;
     }
   else if (_IO_have_backup(fp))
     _IO_free_backup_area(fp);
-  return fp->_jumps->__underflow(fp);
+  return _IO_UNDERFLOW (fp);
+}
+
+int
+DEFUN(__uflow, (fp),
+     _IO_FILE *fp)
+{
+  if (_IO_in_put_mode(fp))
+    if (_IO_switch_to_get_mode(fp) == EOF) return EOF;
+  if (fp->_IO_read_ptr < fp->_IO_read_end)
+    return *(unsigned char*)fp->_IO_read_ptr++;
+  if (_IO_in_backup(fp))
+    {
+      _IO_switch_to_main_get_area(fp);
+      if (fp->_IO_read_ptr < fp->_IO_read_end)
+	return *fp->_IO_read_ptr++;
+    }
+  if (_IO_have_markers(fp))
+    {
+      if (save_for_backup (fp))
+	return EOF;
+    }
+  else if (_IO_have_backup(fp))
+    _IO_free_backup_area(fp);
+  return _IO_UFLOW (fp);
 }
 
 void
-_IO_setb(f, b, eb, a)
-     _IO_FILE *f;
-     char *b;
-     char *eb;
-     int a;
+DEFUN(_IO_setb, (f, b, eb, a),
+      _IO_FILE *f AND char *b AND char *eb AND int a)
 {
   if (f->_IO_buf_base && !(f->_flags & _IO_USER_BUF))
-    FREE_BUF(f->_IO_buf_base);
+    FREE_BUF(f->_IO_buf_base, _IO_blen(f));
   f->_IO_buf_base = b;
   f->_IO_buf_end = eb;
   if (a)
@@ -253,37 +336,45 @@ _IO_setb(f, b, eb, a)
 }
 
 void
-_IO_doallocbuf(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_doallocbuf, (fp),
+      register _IO_FILE *fp)
 {
   if (fp->_IO_buf_base)
     return;
   if (!(fp->_flags & _IO_UNBUFFERED))
-    if (fp->_jumps->__doallocate(fp) != EOF)
+    if (_IO_DOALLOCATE (fp) != EOF)
       return;
   _IO_setb(fp, fp->_shortbuf, fp->_shortbuf+1, 0);
 }
 
 int
-_IO_default_underflow (fp)
-     _IO_FILE *fp;
+DEFUN(_IO_default_underflow, (fp),
+      _IO_FILE *fp)
 {
   return EOF;
 }
 
-_IO_size_t
-_IO_default_xsputn(f, data, n)
-     register _IO_FILE *f;
-     const void *data;
-     _IO_size_t n;
+int
+DEFUN(_IO_default_uflow, (fp),
+      _IO_FILE *fp)
 {
-  register const char *s = data;
+  int ch = _IO_UNDERFLOW (fp);
+  if (ch == EOF)
+    return EOF;
+  return *(unsigned char*)fp->_IO_read_ptr++;
+}
+
+_IO_size_t
+DEFUN(_IO_default_xsputn, (f, data, n),
+      register _IO_FILE *f AND const void *data AND _IO_size_t n)
+{
+  register const char *s = (char*) data;
   register _IO_size_t more = n;
   if (more <= 0)
     return 0;
   for (;;)
     {
-      _IO_size_t count = f->_IO_write_end - f->_IO_write_ptr; /* Space available. */
+      _IO_ssize_t count = f->_IO_write_end - f->_IO_write_ptr; /* Space available. */
       if (count > 0)
 	{
 	  if (count > more)
@@ -313,26 +404,22 @@ _IO_default_xsputn(f, data, n)
 }
 
 _IO_size_t
-_IO_sgetn(fp, data, n)
-     _IO_FILE *fp;
-     void *data;
-     _IO_size_t n;
+DEFUN(_IO_sgetn, (fp, data, n),
+      _IO_FILE *fp AND void *data AND _IO_size_t n)
 {
   /* FIXME handle putback buffer here! */
-  return fp->_jumps->__xsgetn(fp, data, n);
+  return _IO_XSGETN (fp, data, n);
 }
 
 _IO_size_t
-_IO_default_xsgetn(fp, data, n)
-     _IO_FILE *fp;
-     void *data;
-     _IO_size_t n;
+DEFUN(_IO_default_xsgetn, (fp, data, n),
+      _IO_FILE *fp AND void *data AND _IO_size_t n)
 {
   register _IO_size_t more = n;
-  register char *s = data;
+  register char *s = (char*) data;
   for (;;)
     {
-      _IO_size_t count = fp->_IO_read_end - fp->_IO_read_ptr; /* Data available. */
+      _IO_ssize_t count = fp->_IO_read_end - fp->_IO_read_ptr; /* Data available. */
       if (count > 0)
 	{
 	  if (count > more)
@@ -361,22 +448,18 @@ _IO_default_xsgetn(fp, data, n)
 }
 
 int
-_IO_sync(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_sync, (fp),
+      register _IO_FILE *fp)
 {
-  if (fp->_IO_read_ptr == fp->_IO_read_end && fp->_IO_write_ptr == fp->_IO_write_base)
-    return 0;
-  return EOF;
+  return 0;
 }
 
-int
-_IO_default_setbuf(fp, p, len)
-     register _IO_FILE *fp;
-     char* p;
-     _IO_ssize_t len;
+_IO_FILE*
+DEFUN(_IO_default_setbuf, (fp, p, len),
+      register _IO_FILE *fp AND char* p AND _IO_ssize_t len)
 {
-    if (fp->_jumps->__sync(fp) == EOF)
-	return EOF;
+    if (_IO_SYNC (fp) == EOF)
+	return NULL;
     if (p == NULL || len == 0)
       {
 	fp->_flags |= _IO_UNBUFFERED;
@@ -389,33 +472,30 @@ _IO_default_setbuf(fp, p, len)
       }
     fp->_IO_write_base = fp->_IO_write_ptr = fp->_IO_write_end = 0;
     fp->_IO_read_base = fp->_IO_read_ptr = fp->_IO_read_end = 0;
-    return 0;
+    return fp;
 }
 
 _IO_pos_t
-_IO_default_seekpos(fp, pos, mode)
-     _IO_FILE *fp;
-     _IO_pos_t pos;
-     _IO_seekflags mode;
+DEFUN(_IO_default_seekpos, (fp, pos, mode),
+      _IO_FILE *fp AND _IO_pos_t pos AND int mode)
 {
-  return fp->_jumps->__seekoff(fp, _IO_pos_as_off(pos), mode);
+  return _IO_SEEKOFF (fp, _IO_pos_as_off(pos), 0, mode);
 }
 
 int
-_IO_default_doallocate(fp)
-     _IO_FILE *fp;
+DEFUN(_IO_default_doallocate, (fp),
+      _IO_FILE *fp)
 {
-  char *buf = ALLOC_BUF(_IO_BUFSIZ);
-  if (buf == NULL)
-    return EOF;
+  char *buf;
+
+  ALLOC_BUF(buf, _IO_BUFSIZ, EOF);
   _IO_setb(fp, buf, buf+_IO_BUFSIZ, 1);
   return 1;
 }
 
 void
-_IO_init(fp, flags)
-     register _IO_FILE *fp;
-     int flags;
+DEFUN(_IO_init, (fp, flags),
+      register _IO_FILE *fp AND int flags)
 {
   fp->_flags = _IO_MAGIC|flags;
   fp->_IO_buf_base = NULL;
@@ -429,86 +509,108 @@ _IO_init(fp, flags)
   fp->_chain = NULL; /* Not necessary. */
 
   fp->_IO_save_base = NULL;
-  fp->_aux_limit = NULL;
+  fp->_IO_backup_base = NULL;
   fp->_IO_save_end = NULL;
   fp->_markers = NULL;
   fp->_cur_column = 0;
 }
 
 int
-_IO_default_sync (fp)
-     _IO_FILE *fp;
+DEFUN(_IO_default_sync, (fp),
+      _IO_FILE *fp)
 {
   return 0;
 }
 
-void _IO_default_finish (fp)
-     _IO_FILE *fp;
+/* The way the C++ classes are mapped into the C functions in the
+   current implementation, this function can get called twice! */
+
+void
+DEFUN(_IO_default_finish, (fp, dummy),
+      _IO_FILE *fp AND int dummy)
 {
   struct _IO_marker *mark;
   if (fp->_IO_buf_base && !(fp->_flags & _IO_USER_BUF))
-    FREE_BUF(fp->_IO_buf_base);
+    {
+      FREE_BUF(fp->_IO_buf_base, _IO_blen(fp));
+      fp->_IO_buf_base = fp->_IO_buf_end = NULL;
+    }
 
   for (mark = fp->_markers; mark != NULL; mark = mark->_next)
     mark->_sbuf = NULL;
 
-  free (fp->_IO_save_base);
+  if (fp->_IO_save_base)
+    {
+      free (fp->_IO_save_base);
+      fp->_IO_save_base = NULL;
+    }
 
   _IO_un_link(fp);
 }
 
 _IO_pos_t
-_IO_default_seekoff(fp, offset, mode)
-     register _IO_FILE *fp;
-     _IO_off_t offset;
-     _IO_seekflags mode;
+DEFUN(_IO_default_seekoff, (fp, offset, dir, mode),
+      register _IO_FILE *fp AND _IO_off_t offset AND int dir AND int mode)
 {
     return _IO_pos_BAD;
 }
 
 int
-_IO_sputbackc(fp, c) /* ungetc(c, fp) */
-     register _IO_FILE *fp;
-     int c;
+DEFUN(_IO_sputbackc, (fp, c),
+      register _IO_FILE *fp AND int c)
 {
+  int result;
+  
   if (fp->_IO_read_ptr > fp->_IO_read_base
       && (unsigned char)fp->_IO_read_ptr[-1] == (unsigned char)c)
     {
       fp->_IO_read_ptr--;
-      return (unsigned char)c;
+      result = (unsigned char)c;
     }
-  return fp->_jumps->__pbackfail(fp, c);
+  else
+    result = _IO_PBACKFAIL (fp, c);
+
+  if (result != EOF)
+    fp->_flags &= ~_IO_EOF_SEEN;
+
+  return result;
 }
 
 int
-_IO_sungetc(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_sungetc, (fp),
+      register _IO_FILE *fp)
 {
+  int result;
+  
   if (fp->_IO_read_ptr > fp->_IO_read_base)
     {
       fp->_IO_read_ptr--;
-      return (unsigned char)*fp->_IO_read_ptr;
+      result = (unsigned char)*fp->_IO_read_ptr;
     }
   else
-    return fp->_jumps->__pbackfail(fp, EOF);
+    result = _IO_PBACKFAIL (fp, EOF);
+
+  if (result != EOF)
+    fp->_flags &= ~_IO_EOF_SEEN;
+
+  return result;
 }
 
 #if 0 /* Work in progress */
 void
-_IO_set_column(fp, c)
-     register _IO_FILE *fp;
-     int c;
+DEFUN(_IO_set_column, (fp, c),
+      register _IO_FILE *fp AND int c)
 {
   if (c == -1)
-    fp->_collumn = -1;
+    fp->_column = -1;
   else
-    fp->_collumn = c - (fp->_IO_write_ptr - fp->_IO_write_base);
+    fp->_column = c - (fp->_IO_write_ptr - fp->_IO_write_base);
 }
 #else
+
 int
-_IO_set_column(fp, i)
-     register _IO_FILE *fp;
-     int i;
+DEFUN(_IO_set_column, (fp, i),
+      register _IO_FILE *fp AND int i)
 {
   fp->_cur_column = i+1;
   return 0;
@@ -517,10 +619,8 @@ _IO_set_column(fp, i)
 
 
 unsigned
-_IO_adjust_column(start, line, count)
-     unsigned start;
-     const char *line;
-     int count;
+DEFUN(_IO_adjust_column, (start, line, count),
+      unsigned start AND const char *line AND int count)
 {
   register const char *ptr = line + count;
   while (ptr > line)
@@ -530,8 +630,8 @@ _IO_adjust_column(start, line, count)
 }
 
 int
-_IO_get_column(fp)
-     register _IO_FILE *fp;
+DEFUN(_IO_get_column, (fp),
+      register _IO_FILE *fp)
 {
   if (fp->_cur_column) 
     return _IO_adjust_column(fp->_cur_column - 1,
@@ -541,28 +641,56 @@ _IO_get_column(fp)
 }
 
 int
-_IO_flush_all()
+DEFUN_VOID(_IO_flush_all)
 {
   int result = 0;
   _IO_FILE *fp;
+
   for (fp = _IO_list_all; fp != NULL; fp = fp->_chain)
-    if (fp->_jumps->__overflow(fp, EOF) == EOF)
+    if (fp->_IO_write_ptr > fp->_IO_write_base
+	&& _IO_OVERFLOW (fp, EOF) == EOF)
       result = EOF;
   return result;
 }
 
 void
-_IO_flush_all_linebuffered()
+DEFUN_VOID(_IO_flush_all_linebuffered)
 {
   _IO_FILE *fp;
+
   for (fp = _IO_list_all; fp != NULL; fp = fp->_chain)
-    if (fp->_flags & _IO_LINE_BUF)
-      fp->_jumps->__overflow(fp, EOF);
+    if ((fp->_flags & _IO_NO_WRITES) == 0 && fp->_flags & _IO_LINE_BUF)
+      _IO_OVERFLOW (fp, EOF);
 }
 
-void _IO_init_marker(marker, fp)
-     struct _IO_marker *marker;
-     _IO_FILE *fp;
+void
+DEFUN_VOID(_IO_unbuffer_all)
+{
+  _IO_FILE *fp;
+
+  for (fp = _IO_list_all; fp != NULL; fp = fp->_chain)
+    if (! (fp->_flags & _IO_UNBUFFERED))
+      _IO_SETBUF (fp, NULL, 0);
+}
+
+void
+DEFUN_VOID(_IO_cleanup)
+{
+  _IO_flush_all ();
+
+  /* We currently don't have a reliable mechanism for making sure that
+     C++ static destructors are executed in the correct order.
+     So it is possible that other static destructord might want to
+     write to cout - and they're supposed to be able to do so.
+
+     The following will make the standard streambufs be unbuffered, 
+     which forces any output from late destructors to be written out. */
+  _IO_unbuffer_all ();
+}
+
+void
+DEFUN(_IO_init_marker, (marker, fp),
+      struct _IO_marker *marker AND _IO_FILE *fp)
 {
   marker->_sbuf = fp;
   if (_IO_in_put_mode(fp))
@@ -578,8 +706,8 @@ void _IO_init_marker(marker, fp)
 }
 
 void
-_IO_remove_marker (marker)
-     register struct _IO_marker * marker;
+DEFUN(_IO_remove_marker, (marker),
+      register struct _IO_marker *marker)
 {
   /* Unlink from sb's chain. */
   register struct _IO_marker **ptr = &marker->_sbuf->_markers;
@@ -602,17 +730,16 @@ _IO_remove_marker (marker)
 #define BAD_DELTA EOF
 
 int
-_IO_marker_difference(mark1, mark2)
-     struct _IO_marker *mark1;
-     struct _IO_marker *mark2;
+DEFUN(_IO_marker_difference, (mark1, mark2),
+      struct _IO_marker *mark1 AND struct _IO_marker *mark2)
 {
   return mark1->_pos - mark2->_pos;
 }
 
 /* Return difference between MARK and current posistion of MARK's stream. */
 int
-_IO_marker_delta(mark)
-     struct _IO_marker *mark;
+DEFUN(_IO_marker_delta, (mark),
+      struct _IO_marker *mark)
 {
   int cur_pos;
   if (mark->_sbuf == NULL)
@@ -624,10 +751,9 @@ _IO_marker_delta(mark)
   return mark->_pos - cur_pos;
 }
 
-int _IO_seekmark(fp, mark, delta)
-     _IO_FILE *fp;
-     struct _IO_marker *mark;
-     int delta;
+int
+DEFUN(_IO_seekmark, (fp, mark, delta),
+      _IO_FILE *fp AND struct _IO_marker *mark AND int delta)
 {
   if (mark->_sbuf != fp)
     return EOF;
@@ -646,8 +772,9 @@ int _IO_seekmark(fp, mark, delta)
   return 0;
 }
 
-void _IO_unsave_markers(fp)
-     register _IO_FILE *fp;
+void
+DEFUN(_IO_unsave_markers, (fp),
+     register _IO_FILE *fp)
 {
   register struct _IO_marker *mark = fp->_markers;
   if (mark)
@@ -669,13 +796,13 @@ void _IO_unsave_markers(fp)
       fp->_markers = 0;
     }
 
-  _IO_free_backup_area(fp);
+  if (_IO_have_backup(fp))
+    _IO_free_backup_area(fp);
 }
 
 int
-_IO_nobackup_pbackfail(fp, c)
-     register _IO_FILE *fp;
-     int c;
+DEFUN(_IO_nobackup_pbackfail, (fp, c),
+     register _IO_FILE *fp AND int c)
 {
   if (fp->_IO_read_ptr > fp->_IO_read_base)
 	fp->_IO_read_ptr--;
@@ -685,9 +812,8 @@ _IO_nobackup_pbackfail(fp, c)
 }
 
 int
-_IO_default_pbackfail(fp, c)
-     register _IO_FILE *fp;
-     int c;
+DEFUN(_IO_default_pbackfail, (fp, c),
+      register _IO_FILE *fp AND int c)
 {
   if (fp->_IO_read_ptr <= fp->_IO_read_base)
       {
@@ -705,7 +831,7 @@ _IO_default_pbackfail(fp, c)
 	      return EOF;
 	    fp->_IO_save_base = bbuf;
 	    fp->_IO_save_end = fp->_IO_save_base + backup_size;
-	    fp->_aux_limit = fp->_IO_save_end;
+	    fp->_IO_backup_base = fp->_IO_save_end;
 	    _IO_switch_to_backup_area(fp);
 	  }
 	else if (fp->_IO_read_ptr <= fp->_IO_read_base)
@@ -722,7 +848,7 @@ _IO_default_pbackfail(fp, c)
 	    free (fp->_IO_read_base);
 	    _IO_setg(fp,
 		     new_buf, new_buf+(new_size-old_size), new_buf+new_size);
-	    fp->_aux_limit = fp->_IO_read_ptr;
+	    fp->_IO_backup_base = fp->_IO_read_ptr;
 	  }
       }
   fp->_IO_read_ptr--;
@@ -732,36 +858,29 @@ _IO_default_pbackfail(fp, c)
 }
 
 _IO_pos_t
-_IO_default_seek(fp, offset, dir)
-     _IO_FILE *fp;
-     _IO_off_t offset;
-     int dir;
+DEFUN(_IO_default_seek, (fp, offset, dir),
+      _IO_FILE *fp AND _IO_off_t offset AND int dir)
 {
   return _IO_pos_BAD;
 }
 
 int
-_IO_default_stat(fp, st)
-     _IO_FILE *fp;
-     void* st;
+DEFUN(_IO_default_stat, (fp, st),
+      _IO_FILE *fp AND void* st)
 {
   return EOF;
 }
 
 _IO_ssize_t
-_IO_default_read(fp, data, n)
-     register _IO_FILE* fp;
-     void* data;
-     _IO_ssize_t n;
+DEFUN(_IO_default_read, (fp, data, n),
+      register _IO_FILE* fp AND void* data AND _IO_ssize_t n)
 {
   return -1;
 }
 
 _IO_ssize_t
-_IO_default_write(fp, data, n)
-     register _IO_FILE* fp;
-     const void* data;
-     _IO_ssize_t n;
+DEFUN(_IO_default_write, (fp, data, n),
+      register _IO_FILE* fp AND const void* data AND _IO_ssize_t n)
 {
   return 0;
 }
@@ -777,9 +896,121 @@ _IO_default_write(fp, data, n)
 #else
 struct __io_defs {
     __io_defs() { }
-    ~__io_defs() { _IO_flush_all(); }
+    ~__io_defs() { _IO_cleanup(); }
 };   
 __io_defs io_defs__;
 #endif
 
 #endif /* TODO */
+
+#ifdef _IO_USE_LIBC
+/* We only include those files which are not compatible with
+   libio in libc. We need to make sure our version is always
+   used during linking. */
+extern int vfprintf ();
+extern _IO_FILE * popen ();
+static const void * const __dummy [] =
+{
+  &_IO_file_doallocate,
+  &_IO_dtoa,
+  &_IO_strtod,
+  &_IO_adjust_column,
+  &_IO_cleanup,
+  &_IO_default_doallocate,
+  &_IO_default_finish,
+  &_IO_default_pbackfail,
+  &_IO_default_read,
+  &_IO_default_seek,
+  &_IO_default_seekoff,
+  &_IO_default_seekpos,
+  &_IO_default_setbuf,
+  &_IO_default_stat,
+  &_IO_default_sync,
+  &_IO_default_uflow,
+  &_IO_default_underflow,
+  &_IO_default_write,
+  &_IO_default_xsgetn,
+  &_IO_default_xsputn,
+  &_IO_doallocbuf,
+  &_IO_flush_all,
+  &_IO_flush_all_linebuffered,
+  &_IO_free_backup_area,
+  &_IO_get_column,
+  &_IO_init,
+  &_IO_init_marker,
+  &_IO_least_marker,
+  &_IO_link_in,
+  &_IO_marker_delta,
+  &_IO_marker_difference,
+  &_IO_nobackup_pbackfail,
+  &_IO_remove_marker,
+  &_IO_seekmark,
+  &_IO_set_column,
+  &_IO_setb,
+  &_IO_sgetn,
+  &_IO_sputbackc,
+  &_IO_sungetc,
+  &_IO_switch_to_backup_area,
+  &_IO_switch_to_get_mode,
+  &_IO_switch_to_main_get_area,
+  &_IO_sync,
+  &_IO_un_link,
+  &_IO_unbuffer_all,
+  &_IO_unsave_markers,
+  &__overflow,
+  &__uflow,
+  &__underflow,
+  &_IO_do_write,
+  &_IO_file_attach,
+  &_IO_file_close,
+  &_IO_file_close_it,
+  &_IO_file_finish,
+  &_IO_file_fopen,
+  &_IO_file_init,
+  &_IO_file_jumps,
+  &_IO_file_overflow,
+  &_IO_file_read,
+  &_IO_file_seek,
+  &_IO_file_seekoff,
+  &_IO_file_setbuf,
+  &_IO_file_stat,
+  &_IO_file_sync,
+  &_IO_file_underflow,
+  &_IO_file_write,
+  &_IO_file_xsputn,
+  &_IO_vfprintf,
+  &vfprintf,
+  &_IO_vfscanf,
+  &_IO_padn,
+  &_IO_seekoff,
+  &_IO_seekpos,
+  &_IO_outfloat,
+  &_IO_str_count,
+  &_IO_str_finish,
+  &_IO_str_init_readonly,
+  &_IO_str_init_static,
+  &_IO_str_jumps,
+  &_IO_str_overflow,
+  &_IO_str_pbackfail,
+  &_IO_str_seekoff,
+  &_IO_str_underflow,
+  &_IO_fclose,
+  &_IO_popen,
+  &popen,
+  &_IO_proc_close,
+  &_IO_proc_jumps,
+  &_IO_proc_open,
+  &_IO_fdopen,
+  &_IO_fflush,
+  &_IO_fopen,
+  &_IO_fputs,
+  &_IO_fwrite,
+  &_IO_puts,
+  &_IO_setbuffer,
+  &_IO_setvbuf,
+  &_IO_vsprintf,
+  &_IO_vsscanf,
+  &_IO_vasprintf,
+  &_IO_vsnprintf
+};
+#endif
