@@ -68,6 +68,11 @@ static char sccsid[] = "%W% (Berkeley) %G%";
 
 /* end of configuration stuff */
 
+#ifdef __STDC__
+int __cvt_double (double, int, int, int *, int, char *, char *);
+#else
+int __cvt_double ();
+#endif
 
 /*
  * Helper "class" for `fprintf to unbuffered': creates a
@@ -879,3 +884,7 @@ eformat:	if (expcnt) {
 }
 
 #endif /* defined(FLOATING_POINT) && !defined(_IO_USE_DTOA) */
+
+#if defined(__ELF__) && defined(_IO_USE_LIBC) && defined (__GNUC__)
+__asm__(".globl vfprintf; vfprintf  = _IO_vfprintf");
+#endif
